# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: adtzlr

Contique - Numeric continuation of equilibrium equations
Copyright (C) 2021 Andreas Dutzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np

from .helpers import argparser2
from .newtonxt import newtonxt
from . import printinfo


def solve(
    fun,
    x0,
    lpf0,
    jac=None,
    args=(None,),
    dxmax=0.05,
    dlpfmax=0.05,
    control0="lpf",
    jacmode=3,
    jaceps=None,
    maxsteps=50,
    maxcycles=4,
    maxiter=8,
    tol=1e-6,
    overshoot=1.0,
    rebalance=False,
    increase=0.5,
    decrease=2.0,
    high=10,
    low=1e-6,
    minlastfailed=3,
):
    """Numeric continuation of (nonlinear) equilibrium equations.

    Parameters
    ----------
    fun : function
        function in terms of unknows x and optional args which returns the
        equilibrium equations.
    x0 : ndarray
        1d-array with initial values of unknows x
    lpf0 : float
        initial value for the load-proportionality-factor
    jac : function, optional
        jacobian of fun w.r.t. the unknows x
    args : tuple, optional
        Optional tuple of arguments which are passed to the function. Eeven if only
        one argument is passed, it has to be encapsulated in a tuple (default is (None,)).
    dxmax : float, optional
        max. allowed absolute incremental increase of unknowns per step
    dlpfmax : float, optional
        max. allowed absolute incremental increase of lpf per step
    control0 : int, optional
        initial signed control component ( 1-indexed )
    jacmode : int, optional
        forward (2) or central (3) finite-differences approx. of the jacobian
    jaceps : float, optional
        user-specified stepwidth (if None, this defaults to eps^(1/mode))
    maxsteps : int, optional
        max. number of steps
    maxcycles : int, optional
        max. number of cycles per step
    maxiter : int, optional
        max. number of Newton-iterations per cycle
    tol : float, optional
        tolerated residual of the norm of the equilibrium equation (default is 1e-8)
    overshoot : float, optional
        allowed overshoot of the final control component of a cycle (default is 1.0)
    rebalance : bool, optional
        rebalance max. allowed incremental increase values after each step
    increase : float, optional
        rebalance increase factor
    decrease : float, optional
        rebalance decrease factor
    high : float, optional
        rebalance max. factor of incremental increase w.r.t. to the initial values
    low : float, optional
        rebalance min. factor of incremental increase w.r.t. to the initial values
    minlastfailed : int, optional
        rebalance increase only after a given number of converged steps

    Returns
    -------
    Res : list
        List of NewtonResults (with res.x being the final unknowns per step)

    """

    # allow passing empty *args to fun(x, lpf)
    fun = argparser2(fun)

    # init number of unknows
    ncomp = 1 + len(x0)

    # init rebalance and lastfailed
    # (not used if not rebalance)
    rebalanced = False
    lastfailed = 0

    # initial control component
    if control0 == "lpf":
        j0 = ncomp
    else:
        j0 = control0

    # init y=(x,l)-combined quantities
    y0 = np.append(x0, lpf0)
    dymax = np.append(np.ones_like(x0) * dxmax, dlpfmax)
    dymax0 = dymax.copy()

    # init list of results
    Res = [newtonxt(fun, jac, y0, j0, dymax, jacmode, jaceps, args, maxiter=0, tol=tol)]

    printinfo.header()

    # Step loop.
    for step in 1 + np.arange(maxsteps):
        ## pre-identification of control component
        res = newtonxt(
            fun, jac, y0, j0, dymax, jacmode, jaceps, args, maxiter=1, tol=tol
        )

        # Cycle loop.
        for cycl in 1 + np.arange(maxcycles):

            # Newton Iterations.
            res = newtonxt(
                fun, jac, y0, j0, dymax, jacmode, jaceps, args, maxiter=maxiter, tol=tol
            )
            printinfo.cycle(
                step,
                cycl,
                j0,
                res.control,
                res.status,
                np.linalg.norm(res.fun),
                res.niterations,
                max(abs(res.dys)) <= overshoot,
            )

            # Did Newton Iterations converge?
            if res.success:

                # Did control component change? OR Was overshoot inside allowed range?
                if (res.control == j0) or (max(abs(res.dys)) <= overshoot):

                    # Save results, move to next step.
                    j0 = res.control
                    y0 = res.x
                    Res.append(res)
                    break

                else:  # Were max. number of cycles reached?
                    if cycl == maxcycles:
                        # Print Error and set success-flag to False.
                        printinfo.errorcontrol()
                        res.success = False
                    else:
                        # re-cycle Step with new control component
                        j0 = res.control
            else:
                # break cycle loop if Newton Iterations failed.
                break

        # Rebalance max. incremental unknowns
        # --------------------------------------------------------------------
        if rebalance:
            dymaxn = dymax.copy()
            dymax, rebalanced, lastfailed = adjust(
                dymax0,
                dymaxn,
                success=res.success,
                n=res.niterations,
                lastfailed=lastfailed,
                increase=increase,
                decrease=decrease,
                high=high,
                low=low,
                minlastfailed=minlastfailed,
                nref=8,
            )
        # --------------------------------------------------------------------

        # break step loop if Newton Iterations failed.
        if not res.success and not rebalanced:
            printinfo.errorfinal()
            break

    return Res


def adjust(
    x0,
    xn,
    success,
    n,
    lastfailed,
    increase=0.5,
    decrease=4.0,
    high=10,
    low=1e-6,
    minlastfailed=3,
    nref=8,
):
    "Adjust (rebalance) the stepwidth. **Warning**: Experimental feature."

    rebalanced = True
    if success:
        lastfailed += 1
        if lastfailed >= minlastfailed:
            x = xn * (1 + (nref - min(n, nref)) / nref * increase)
        else:  # lastfailed < minlastfailed
            x = xn

    elif not success:
        x = xn / decrease
        lastfailed = 0

    y = np.maximum(np.minimum(x / x0, high), low) * x0
    if y[0] == xn[0]:
        rebalanced = False

    return y, rebalanced, lastfailed



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: adtzlr

Contique - Numeric continuation of equilibrium equations
Copyright (C) 2021 Andreas Dutzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from .helpers import argparser


class NewtonResult:
    """Class for handling the results of a Newton-Rhapson solution.

    This class has several public attribues.

    Attributes
    ----------
    success : bool
        flag for the converged solution
    message : str
        message for the state
    status : int
        integer representig the status of the solution (0 if not converged, 1 if converged).
    niterations : int
        number of performed iterations
    x : ndarray
        1d-array containing the unknows
    fun : function
        function returning the equilibrium equations
    jac : function
        function returning the jacobian of the equilibrium equations

    """

    def __init__(self, fun, x0, jac, args):
        """Initialize an Instance of a NewtonResult.

        Parameters
        ----------
        fun : function
            function returning the equilibrium equations
        x0 : ndarray
            1d-array containing the initial unknows
        jac : function
            function returning the jacobian of the equilibrium equations
        args : tuple, optional
            Optional tuple of arguments which are passed to the function. Eeven if only
            one argument is passed, it has to be encapsulated in a tuple (default is (None,)).

        """
        self.success = False
        self.message = "not started"
        self.status = 0
        self.niterations = 0
        self.x = x0.copy()
        self.fun = argparser(fun)(self.x, *args)
        self.jac = argparser(jac)(self.x, *args)


def newtonrhapson(fun, x0, jac, args=(None,), maxiter=8, tol=1e-8):
    """A simple n-dimensional Newton-Rhapson solver.

    Parameters
    ----------
    fun : function
        function in terms of unknows x and optional args which returns the
        equilibrium equations.
    x0 : ndarray
        1d-array with initial values of unknows x
    jac : function
        jacobian of fun w.r.t. the unknows x
    args : tuple, optional
        Optional tuple of arguments which are passed to the function. Eeven if only
        one argument is passed, it has to be encapsulated in a tuple (default is (None,)).
    maxiter : int, optional
        maximum number of iterations (default is 8)
    tol : float, optional
        tolerated residual of the norm of the equilibrium equation (default is 1e-8)

    Returns
    -------
    res : NewtonResult
        Instance of NewtonResult with res.x being the final unknowns

    """

    # init result object with initial function evaluation
    res = NewtonResult(fun, x0, jac, args)

    # iteration loop
    for res.niterations in range(1, 1 + maxiter):

        # solve linear equation system
        res.x += np.linalg.solve(res.jac, -res.fun)

        # calculate function and jacobian at updated x
        res.fun = argparser(fun)(res.x, *args)
        res.jac = argparser(jac)(res.x, *args)

        # convergence check
        if np.linalg.norm(res.fun) < tol:
            res.success = True
            res.status = 1
            res.message = "Solution converged in {0:2d} Iteration".format(
                res.niterations
            )
            if res.niterations > 1:
                res.message = res.message + "s"

            break

    # check if newton process failed
    if not res.success:
        if maxiter == 1:
            res.message = "Calculated linear solution because of input parameter `maxiter=1` (not converged)."
        else:
            res.message = "Newton-R. process failed."

    return res



import matplotlib.pyplot as plt
import numpy as np
import pytest

import contique


def fun(x, lpf):
    n = len(x)
    h = 1 / (n - 1)
    A = np.diag(2 * np.ones_like(x) / h ** 2)
    for i in [1, -1]:
        A -= np.diag(np.ones_like(x[:-1]) / h ** 2, i)
    f = -A.dot(x) + lpf * np.exp(x)
    for i in [0, -1]:
        f[i] = x[i]
    return f


def val(x):
    n = len(x)
    h = 1 / (n - 1)
    H = np.ones(n) * h
    for i in [0, -1]:
        H[i] = h / 2
    return np.sqrt(x.dot(H * x))


def test_bratu():

    # initial solution
    n = 51
    lpf0 = 0.0
    x0 = np.zeros(n)

    # numeric continuation
    # + use rebalance step-width
    # + don't wait until some steps converged until rebalance-increase
    Res = contique.solve(
        fun=fun,
        x0=x0,
        lpf0=lpf0,
        dxmax=0.5,
        dlpfmax=0.5,
        maxsteps=22,
        tol=1e-10,
        rebalance=True,
        minlastfailed=0,
    )

    # extract/calculate load-proportionality factor and unknown u
    lpf = np.array([res.x[-1] for res in Res])
    u = np.array([val(res.x[:-1]) for res in Res])

    # create figure
    plt.plot(lpf, u, "x-")
    plt.xlabel("load-proportionality-factor $LPF$")
    plt.ylabel("$||u||_2$")
    plt.plot(0, 0, "C0o", ms=10)
    plt.savefig("test_bratu.svg")


if __name__ == "__main__":
    test_bratu()



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: adtzlr

Contique - Numeric continuation of equilibrium equations
Copyright (C) 2021 Andreas Dutzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from .jacobian import jacobian
from .helpers import needle, control
from .newton import newtonrhapson


def funxt(y, n, ymax, fun, jac, jacmode, jaceps, args):
    """Extended equilibrium equations.

    Parameters
    ----------
    y : array
        1d-array of unknowns.
    n : array
        pre-evaluated needle-vector.
    ymax : array
        1d-array with max. allowed values of unknows.
    fun : function
        1d-array of equilibrium equations
    jac : function
        jacobian of equilibrium euqations (not used)
    jacmode : int
        2 or 3-point evaulation of the jacobian (not used)
    jaceps : float
        a small number to evaulate the jacobian (not used)
    args : tuple
        optional function arguments

    Returns
    -------
    array
        extended 1d-array of equilibrium equations
        with control equation
    """

    x, l = y[:-1], y[-1]
    return np.append(fun(x, l, *args), np.dot(n, (y - ymax)))


def jacxt(y, n, ymax, fun, jac=None, jacmode=3, jaceps=None, args=(None,)):
    """Jacobian of extended equilibrium equations.

    Parameters
    ----------
    y : ndarray
        1d-array of extended unknows
    n : ndarray
        1d-array with pre-evaluated needle-vector
    ymax : ndarray
        1d-array with max. allowed incremental increase values of y
    fun : function
        function in terms of unknows x and optional args which returns the
        equilibrium equations.
    jac : function, optional
        jacobian of fun w.r.t. the extended unknows y
    jacmode : int, optional
        forward (2) or central (3) finite-differences approx. of the jacobian
    jaceps : float, optional
        user-specified stepwidth (if None, this defaults to eps^(1/mode))
    args : tuple, optional
        Optional tuple of arguments which are passed to the function. Eeven if only
        one argument is passed, it has to be encapsulated in a tuple (default is (None,)).

    Returns
    -------
        ndarray
        jacobian of fun w.r.t. y (contains both derivatives of x and lpf) as 2d-array
    """
    x, lpf = y[:-1], y[-1]
    if jac is None:
        dfundx = jacobian(fun, argnum=0, mode=jacmode, h=jaceps)
        dfundl = jacobian(fun, argnum=1, mode=jacmode, h=jaceps)
    else:
        dfundx, dfundl = jac
    return np.vstack(
        (np.hstack((dfundx(x, lpf, *args), dfundl(x, lpf, *args).reshape(-1, 1))), n)
    )


def newtonxt(
    fun,
    jac,
    y0,
    control0,
    dymax,
    jacmode=3,
    jaceps=None,
    args=(None,),
    maxiter=20,
    tol=1e-8,
):
    """Solve equilibrium equations starting from an initial solution
    with control component and max. allowed increase of unknowns.

    Parameters
    ----------
    fun : function
        function in terms of unknows x and optional args which returns the
        equilibrium equations.
    jac : function, optional
        jacobian of fun w.r.t. the extended unknows y
    y0 : ndarray
        1d-array of initial extended unknows
    control0 : int, optional
        initial signed control component ( 1-indexed )
    dxmax : float, optional
        max. allowed absolute incremental increase of extended unknowns per step
    jacmode : int, optional
        forward (2) or central (3) finite-differences approx. of the jacobian
    jaceps : float, optional
        user-specified stepwidth (if None, this defaults to eps^(1/mode))
    args : tuple, optional
        Optional tuple of arguments which are passed to the function. Eeven if only
        one argument is passed, it has to be encapsulated in a tuple (default is (None,)).
    maxiter : int, optional
        max. number of Newton-iterations per cycle
    tol : float, optional
        tolerated residual of the norm of the equilibrium equation (default is 1e-8)

    Returns
    -------
    res : NewtonResult
        Instance of NewtonResult with res.x being the final extended unknowns
    """

    # init needle-vector and obtain ymax
    n = needle(control0, len(y0))
    ymax = y0 + np.sign(control0) * dymax

    # Newton-Rhapson solver
    res = newtonrhapson(
        fun=funxt,
        x0=y0,
        jac=jacxt,
        args=(n, ymax, fun, jac, jacmode, jaceps, args),
        maxiter=maxiter,
        tol=tol,
    )

    # normalized dy = dy/dymax
    res.dys = (res.x - y0) / dymax

    # final control component based on dnormalized dy
    res.control = control(res.dys)

    return res



import matplotlib.pyplot as plt
import numpy as np
import pytest

import contique


def fun(x, l, a, b):
    return np.array([-a * np.sin(x[0]) + x[1] ** 2 + l, -b * np.cos(x[1]) * x[1] + l])


def test_sincos():

    # initial solution
    x0 = np.zeros(2)
    lpf0 = 0.0

    # additional function arguments
    a, b = 1, 1

    # numeric continuation
    Res = contique.solve(
        fun=fun,
        x0=x0,
        args=(a, b),
        lpf0=lpf0,
        dxmax=0.1,
        dlpfmax=0.1,
        maxsteps=75,
        maxcycles=4,
        maxiter=20,
        tol=1e-10,
        overshoot=1.0,
    )

    X = np.array([res.x for res in Res])

    plt.plot(X[:, 0], X[:, 1], "C0.-")
    plt.xlabel("$x_1$")
    plt.ylabel("$x_2$")
    plt.plot([0], [0], "C0o", lw=3)
    plt.arrow(
        X[-2, 0],
        X[-2, 1],
        X[-1, 0] - X[-2, 0],
        X[-1, 1] - X[-2, 1],
        head_width=0.075,
        head_length=0.15,
        fc="C0",
        ec="C0",
    )
    plt.gca().set_aspect("equal")
    plt.savefig("test_sincos.svg")


if __name__ == "__main__":
    test_sincos()



import matplotlib.pyplot as plt
import numpy as np
import pytest

import contique


def fun(x, l, a):
    r = a / np.sqrt(l)
    return np.array([-x[0] + r * np.cos(l), -x[1] + r * np.sin(l)])


def test_lituus_spiral():

    # additional function arguments
    a = 1

    # initial solution
    lpf0 = 0.2
    x0 = fun(np.zeros(2), lpf0, a)

    # lpf0 = 1.0
    # x0 = np.array([1.,0.])

    # numeric continuation
    Res = contique.solve(
        fun=fun,
        x0=x0,
        args=(a,),
        lpf0=lpf0,
        control0=3,
        dxmax=0.2,
        dlpfmax=0.2,
        jacmode=3,
        jaceps=1e-4,
        maxsteps=500,
        maxcycles=4,
        maxiter=20,
        tol=1e-12,
        overshoot=1.05,
    )

    X = np.array([res.x for res in Res])

    plt.plot(X[:, 0], X[:, 1], "-")
    plt.xlabel("$x_1$")
    plt.ylabel("$x_2$")
    plt.xlim(-3, 3)
    plt.ylim(-3, 3)
    plt.plot(X[0, 0], X[0, 1], "C0o", lw=3)
    # plt.arrow(X[-2,0],X[-2,1],X[-1,0]-X[-2,0],X[-1,1]-X[-2,1],
    #          head_width=0.1, head_length=0.2, fc='C0', ec='C0')
    plt.gca().set_aspect("equal")
    plt.savefig("test_lituus_spiral.svg")


if __name__ == "__main__":
    test_lituus_spiral()



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: adtzlr

Contique - Numeric continuation of equilibrium equations
Copyright (C) 2021 Andreas Dutzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np


def needle(j, m):
    """Return an array of zeros with a needle at component `j` and length `m`.
    This array is used to slice out the j-th component of another 1d-array in
    `xj = needle(j,m).dot(x)`. Furthermore, the derivative of this equation
    w.r.t. `x` results in `needle(j,m)`.

    Parameters
    ----------
    j : int
        Signed 1-indexed (!) component at which the needle should be positioned.
    m : int
        Length of the output array.

    Returns
    -------
    ndarray
        1d-array of length with a needle at item j and zero for all other items.

    Examples
    --------
    A needle vector with length m=9 and a needle at j=5.

    >>> needle(-6,9)
    array([0, 0, 0, 0, 0, 1, 0, 0, 0])

    """

    # init a 1d-array filled with zeros
    n = np.zeros(m, dtype=int)

    # insert needle at position `j`
    n[abs(j) - 1] = 1

    return n


def control(x):
    """Obtain the signed index of the greatest absolute value of a
    1d-array. The returned integer is 1-indexed (!) and multiplied
    by the sign of the greatest value.

    Parameters
    ----------
    x : ndarray
        1d-array

    Returns
    -------
    int
        1-indexed (!) position of which contains the greatest absolute value
        of `x` multiplied by the sign of this greatest value.
    """

    # 0-indexed position
    j = abs(x).argmax()

    # convert to 1-indexed position and multiply by the sign of `x[j]`
    return int((j + 1) * np.sign(x[j]))


def argparser(fun):
    "Function decorator for the handling of function arguments."

    def inner(x, *args, **kwargs):
        "Pass \*args and \*\*kwargs to a function if they are not None."
        no_args = (len(args) == 1 and args[0] is None) or not bool(args)
        no_kwargs = not bool(kwargs)

        if no_args and no_kwargs:
            f = fun(x)
        elif not no_args and no_kwargs:
            f = fun(x, *args)
        elif no_args and not no_kwargs:
            f = fun(x, **kwargs)
        else:
            f = fun(x, *args, **kwargs)
        return f

    return inner


def argparser2(fun):
    """Function decorator for the handling of function arguments
    with 2 primary arguments followed by other args."""

    def inner2(x, lpf, *args, **kwargs):
        "Pass \*args and \*\*kwargs to a function if they are not None."
        no_args = (len(args) == 1 and args[0] is None) or not bool(args)
        no_kwargs = not bool(kwargs)

        if no_args and no_kwargs:
            f = fun(x, lpf)
        elif not no_args and no_kwargs:
            f = fun(x, lpf, *args)
        elif no_args and not no_kwargs:
            f = fun(x, lpf, **kwargs)
        else:
            f = fun(x, lpf, *args, **kwargs)
        return f

    return inner2



import matplotlib.pyplot as plt
import numpy as np
import pytest

import contique


def fun(x, l):
    a = 1
    r = a * l
    return np.array([-x[0] + r * np.cos(l), -x[1] + r * np.sin(l)])


def test_archimedian_spiral_noargs():

    # initial solution
    x0 = np.zeros(2)
    lpf0 = 0.0

    # numeric continuation
    Res = contique.solve(
        fun=fun,
        x0=x0,
        lpf0=lpf0,
        dxmax=0.2,
        dlpfmax=0.2,
        maxsteps=500,
        maxcycles=4,
        maxiter=8,
        tol=1e-10,
        overshoot=1.05,
    )

    X = np.array([res.x for res in Res])

    plt.plot(X[:, 0], X[:, 1], "-")
    plt.xlabel("$x_1$")
    plt.ylabel("$x_2$")
    plt.xlim(-15, 15)
    plt.ylim(-15, 15)
    plt.plot([0], [0], "C0o", lw=3)
    plt.arrow(
        X[-2, 0],
        X[-2, 1],
        X[-1, 0] - X[-2, 0],
        X[-1, 1] - X[-2, 1],
        head_width=1,
        head_length=2,
        fc="C0",
        ec="C0",
    )
    plt.gca().set_aspect("equal")
    plt.savefig("test_archimedean_spiral_noargs.svg")


if __name__ == "__main__":
    test_archimedian_spiral_noargs()



import matplotlib.pyplot as plt
import numpy as np
import pytest

import contique


def fun(x, lpf, a, L, EA):
    WL = -x[0] / L
    lL = np.sqrt(1 - 2 * np.sin(a) * WL + WL ** 2)
    N = EA * (lL - 1)
    return np.array([2 * N * (np.sin(a) - WL) + lpf])


def test_twotruss():

    # initial solution
    x0 = np.zeros(1)
    lpf0 = 0.0

    # args
    L = np.sqrt(2)
    a = np.deg2rad(45)
    EA = 1

    # numeric continuation
    Res = contique.solve(fun=fun, x0=x0, lpf0=lpf0, args=(a, L, EA))

    X = np.array([res.x for res in Res])
    print(X[-2:, :])

    plt.plot(X[:, 0], X[:, 1], ".-")
    plt.xlabel("$W$")
    plt.ylabel("$LPF$")
    plt.plot([0], [0], "C0o", lw=3)
    plt.arrow(
        X[-2, 0],
        X[-2, 1],
        X[-1, 0] - X[-2, 0],
        X[-1, 1] - X[-2, 1],
        head_width=0.05,
        head_length=0.1,
        fc="C0",
        ec="C0",
    )
    plt.gca().set_aspect("equal")
    plt.savefig("test_twotruss.svg")


if __name__ == "__main__":
    test_twotruss()



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: adtzlr

Contique - Numeric continuation of equilibrium equations
Copyright (C) 2021 Andreas Dutzler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np


def header():
    print("|Step,C.| Control Comp. | Norm (Iter.#) | Message     |")
    print("|-------|---------------|---------------|-------------|")


def cycle(step, cycl, j0, control, status, fnorm, niterations, overshootcond):
    if cycl > 1:
        stp = "     "
    else:
        stp = "{0:4d},".format(step)

    if j0 != control and status == 1:
        if overshootcond:
            sts = 3
        else:
            sts = 2
    else:
        sts = status

    message = ["Failed       ", " " * 13, " => re-Cycle ", "tol.Overshoot"]
    print(
        "|{0:4s}{1:1d} | {2:+4d}  => {3:+4d} | {4:.1e} ({5:2d}#) |{6:13s}|".format(
            stp, cycl, j0, control, fnorm, niterations, message[sts]
        )
    )


def errorcontrol():
    print("")
    print("ERROR. Control component changed in last cycle.")
    print("       Possible solution: Reduce stepwidth.")


def errorfinal():
    print("")
    print("ERROR. Numerical continuation stopped.")
